import cupy as cp
from pygpe.spinone.wavefunction import Wavefunction


def step_wavefunction(wfn: Wavefunction, params: dict) -> None:
    """Propagates the wavefunction forward one time step.

    :param wfn: The wavefunction of the system.
    :type wfn: :class:`Wavefunction`
    :param params: The parameters of the system.
    :type params: dict
    """
    _kinetic_zeeman_step(wfn, params)
    wfn.ifft()
    _interaction_step(wfn, params)
    wfn.fft()
    _kinetic_zeeman_step(wfn, params)
    if isinstance(params["dt"], complex):
        _renormalise_wavefunction(wfn)


def _kinetic_zeeman_step(wfn: Wavefunction, pm: dict) -> None:
    """Computes the kinetic-zeeman subsystem for half a time step.

    :param wfn: The wavefunction of the system.
    :param pm: The parameter. dictionary.
    """
    wfn.fourier_plus_component *= cp.exp(
        -0.25 * 1j * pm["dt"] * (wfn.grid.wave_number + 2 * pm["q"])
    )
    wfn.fourier_zero_component *= cp.exp(-0.25 * 1j * pm["dt"] * wfn.grid.wave_number)
    wfn.fourier_minus_component *= cp.exp(
        -0.25 * 1j * pm["dt"] * (wfn.grid.wave_number + 2 * pm["q"])
    )


def _interaction_step(wfn: Wavefunction, pm: dict) -> None:
    """Computes the interaction subsystem for a full time step.

    :param wfn: The wavefunction of the system.
    :param pm: The parameters' dictionary.
    """
    spin_perp, spin_z = _calculate_spins(wfn)
    spin_mag = cp.sqrt(abs(spin_perp) ** 2 + spin_z**2)
    dens = _calculate_density(wfn)

    # Trig terms needed in solution
    cos_term = cp.cos(pm["c2"] * spin_mag * pm["dt"])
    sin_term = cp.nan_to_num(1j * cp.sin(pm["c2"] * spin_mag * pm["dt"]))

    plus_comp_temp = cos_term * wfn.plus_component - sin_term * (
        spin_z * wfn.plus_component
        + cp.conj(spin_perp) / cp.sqrt(2) * wfn.zero_component
    )
    zero_comp_temp = cos_term * wfn.zero_component - sin_term / cp.sqrt(2) * (
        spin_perp * wfn.plus_component + cp.conj(spin_perp) * wfn.minus_component
    )
    minus_comp_temp = cos_term * wfn.minus_component - sin_term * (
        spin_perp / cp.sqrt(2) * wfn.zero_component - spin_z * wfn.minus_component
    )

    wfn.plus_component = plus_comp_temp * cp.exp(
        -1j * pm["dt"] * (pm["trap"] - pm["p"] + pm["c0"] * dens)
    )
    wfn.zero_component = zero_comp_temp * cp.exp(
        -1j * pm["dt"] * (pm["trap"] + pm["c0"] * dens)
    )
    wfn.minus_component = minus_comp_temp * cp.exp(
        -1j * pm["dt"] * (pm["trap"] + pm["p"] + pm["c0"] * dens)
    )


def _calculate_spins(wfn: Wavefunction) -> tuple[cp.ndarray, cp.ndarray]:
    """Calculates the perpendicular and longitudinal spins.

    :param wfn: The wavefunction of the system.
    :return: The perpendicular & longitudinal spin, respectively.
    """
    spin_perp = cp.sqrt(2.0) * (
        cp.conj(wfn.plus_component) * wfn.zero_component
        + cp.conj(wfn.zero_component) * wfn.minus_component
    )
    spin_z = cp.abs(wfn.plus_component) ** 2 - cp.abs(wfn.minus_component) ** 2

    return spin_perp, spin_z


def _calculate_density(wfn: Wavefunction) -> cp.ndarray:
    """Calculates the total condensate density.

    :param wfn: The wavefunction of the system.
    :return: The total atomic density.
    """
    return (
        cp.abs(wfn.plus_component) ** 2
        + cp.abs(wfn.zero_component) ** 2
        + cp.abs(wfn.minus_component) ** 2
    )


def _renormalise_wavefunction(wfn: Wavefunction) -> None:
    """Re-normalises the wavefunction to the correct atom number.

    :param wfn: The wavefunction of the system.
    """
    wfn.ifft()
    correct_atom_plus, correct_atom_zero, correct_atom_minus = (
        wfn.atom_num_plus,
        wfn.atom_num_zero,
        wfn.atom_num_minus,
    )
    current_atom_plus, current_atom_zero, current_atom_minus = _calculate_atom_num(wfn)
    wfn.plus_component *= cp.sqrt(correct_atom_plus / current_atom_plus)
    wfn.zero_component *= cp.sqrt(correct_atom_zero / current_atom_zero)
    wfn.minus_component *= cp.sqrt(correct_atom_minus / current_atom_minus)
    wfn.fft()


def _calculate_atom_num(wfn: Wavefunction) -> tuple[int, int, int]:
    """Calculates the atom number of each wavefunction component.

    :param wfn: The wavefunction of the system.
    :return: The atom numbers of the plus, zero, and minus components, respectively.
    """
    atom_num_plus = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.plus_component) ** 2
    )
    atom_num_zero = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.zero_component) ** 2
    )
    atom_num_minus = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.minus_component) ** 2
    )

    return atom_num_plus, atom_num_zero, atom_num_minus



import h5py
import cupy as cp
from pygpe.shared.grid import Grid
from pygpe.shared import data_manager_paths as dmp
from pygpe.spinhalf.wavefunction import Wavefunction


class DataManager:
    """This object handles all the data of the simulation, including the wavefunction, grid, and parameter data.

    :param filename: The name of the data file.
    :type filename: str
    :param data_path: The relative path to the folder containing the data file.
    :type data_path: str

    :ivar filename: The name of the data file.
    :ivar data_path: The relative path to the folder containing the data file.
    """

    def __init__(self, filename: str, data_path: str):
        self.filename = filename
        self.data_path = data_path
        self._time_index = 0

        # Create file
        h5py.File(f"{self.data_path}/{self.filename}", "w")

    def save_initial_parameters(
        self, grid: Grid, wfn: Wavefunction, parameters: dict
    ) -> None:
        """Saves the initial grid, wavefunction and parameters to a HDF5 file.

        :param grid: The grid object of the system.
        :type grid: :class:`Grid`
        :param wfn: The wavefunction of the system.
        :type wfn: :class:`Wavefunction`
        :param parameters: The parameter dictionary.
        :type parameters: dict
        """
        self._save_initial_grid_params(grid)
        self._save_initial_wfn(wfn)
        self._save_params(parameters)

    def _save_initial_grid_params(self, grid: Grid) -> None:
        """Creates new datasets in file for grid-related parameters and saves
        initial values.
        """
        if grid.ndim == 1:
            with h5py.File(f"{self.data_path}/{self.filename}", "r+") as file:
                file.create_dataset(dmp.GRID_NX, data=grid.num_points_x)
                file.create_dataset(dmp.GRID_DX, data=grid.grid_spacing_x)
        elif grid.ndim == 2:
            with h5py.File(f"{self.data_path}/{self.filename}", "r+") as file:
                file.create_dataset(dmp.GRID_NX, data=grid.num_points_x)
                file.create_dataset(dmp.GRID_NY, data=grid.num_points_y)
                file.create_dataset(dmp.GRID_DX, data=grid.grid_spacing_x)
                file.create_dataset(dmp.GRID_DY, data=grid.grid_spacing_y)
        elif grid.ndim == 3:
            with h5py.File(f"{self.data_path}/{self.filename}", "r+") as file:
                file.create_dataset(dmp.GRID_NX, data=grid.num_points_x)
                file.create_dataset(dmp.GRID_NY, data=grid.num_points_y)
                file.create_dataset(dmp.GRID_NZ, data=grid.num_points_z)
                file.create_dataset(dmp.GRID_DX, data=grid.grid_spacing_x)
                file.create_dataset(dmp.GRID_DY, data=grid.grid_spacing_y)
                file.create_dataset(dmp.GRID_DZ, data=grid.grid_spacing_z)

    def _save_initial_wfn(self, wfn: Wavefunction) -> None:
        """Creates new datasets in file for the wavefunction and saves
        initial values.
        """
        if wfn.grid.ndim == 1:
            with h5py.File(f"{self.data_path}/{self.filename}", "r+") as file:
                file.create_dataset(
                    dmp.SPINHALF_WAVEFUNCTION_PLUS,
                    (wfn.grid.shape, 1),
                    maxshape=(wfn.grid.shape, None),
                    dtype="complex128",
                )
                file.create_dataset(
                    dmp.SPINHALF_WAVEFUNCTION_MINUS,
                    (wfn.grid.shape, 1),
                    maxshape=(wfn.grid.shape, None),
                    dtype="complex128",
                )
        else:
            with h5py.File(f"{self.data_path}/{self.filename}", "r+") as file:
                file.create_dataset(
                    dmp.SPINHALF_WAVEFUNCTION_PLUS,
                    (*wfn.grid.shape, 1),
                    maxshape=(*wfn.grid.shape, None),
                    dtype="complex128",
                )
                file.create_dataset(
                    dmp.SPINHALF_WAVEFUNCTION_MINUS,
                    (*wfn.grid.shape, 1),
                    maxshape=(*wfn.grid.shape, None),
                    dtype="complex128",
                )

    def _save_params(self, parameters: dict) -> None:
        """Creates new datasets in file for condensate & time parameters
        and saves initial values.
        """
        with h5py.File(f"{self.data_path}/{self.filename}", "r+") as file:
            for key in parameters:
                file.create_dataset(f"parameters/{key}", data=parameters[key])

    def save_wavefunction(self, wfn: Wavefunction) -> None:
        """Saves the current wavefunction data to the dataset.

        :param wfn: The wavefunction of the system.
        :type wfn: :class:`Wavefunction`
        """
        wfn.ifft()  # Update real-space wavefunction before saving
        if wfn.grid.ndim == 1:
            with h5py.File(f"{self.data_path}/{self.filename}", "r+") as data:
                new_psi_plus = data[dmp.SPINHALF_WAVEFUNCTION_PLUS]
                new_psi_plus.resize((wfn.grid.num_points_x, self._time_index + 1))
                new_psi_plus[:, self._time_index] = cp.asnumpy(wfn.plus_component)

                new_psi_minus = data[dmp.SPINHALF_WAVEFUNCTION_MINUS]
                new_psi_minus.resize((wfn.grid.num_points_x, self._time_index + 1))
                new_psi_minus[:, self._time_index] = cp.asnumpy(wfn.minus_component)
        else:
            with h5py.File(f"{self.data_path}/{self.filename}", "r+") as data:
                new_psi_plus = data[dmp.SPINHALF_WAVEFUNCTION_PLUS]
                new_psi_plus.resize((*wfn.grid.shape, self._time_index + 1))
                new_psi_plus[..., self._time_index] = cp.asnumpy(wfn.plus_component)

                new_psi_minus = data[dmp.SPINHALF_WAVEFUNCTION_MINUS]
                new_psi_minus.resize((*wfn.grid.shape, self._time_index + 1))
                new_psi_minus[..., self._time_index] = cp.asnumpy(wfn.minus_component)

        self._time_index += 1



from pygpe.shared.grid import Grid
import cupy as cp


class Wavefunction:
    def __init__(self, grid: Grid):
        """Constructs the wavefunction object."""
        self.grid = grid

        self.plus_component = cp.empty(grid.shape, dtype="complex128")
        self.minus_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_plus_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_minus_component = cp.empty(grid.shape, dtype="complex128")

        self.atom_num_plus = 0
        self.atom_num_minus = 0

    def set_wavefunction_components(
        self, plus_component: cp.ndarray, minus_component: cp.ndarray
    ) -> None:
        """Set the wavefunction components to the specified arrays.

        :param plus_component: Plus component array.
        :param minus_component: Minus component array.
        """
        self.plus_component = plus_component
        self.minus_component = minus_component

        self.fft()
        self._update_atom_numbers()

    def _update_atom_numbers(self) -> None:
        """Updates atom number variables after change in wavefunction."""
        self.atom_num_plus = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.plus_component) ** 2
        )
        self.atom_num_minus = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.minus_component) ** 2
        )

    def add_noise_to_components(
        self, components: str, mean: float, std_dev: float
    ) -> None:
        """Adds noise to the specified wavefunction components
        using a normal distribution.

        :param components: "plus", "minus", or "all": The components to add noise to.
        :param mean: The mean of the normal distribution.
        :param std_dev: The standard deviation of the normal distribution.
        """
        match components.lower():
            case "plus":
                self.plus_component += self._generate_complex_normal_dist(mean, std_dev)
            case "minus":
                self.minus_component += self._generate_complex_normal_dist(
                    mean, std_dev
                )
            case "all":
                self.plus_component += self._generate_complex_normal_dist(mean, std_dev)
                self.minus_component += self._generate_complex_normal_dist(
                    mean, std_dev
                )
            case _:
                raise ValueError(f"{components} is not a supported configuration")

    def _generate_complex_normal_dist(self, mean: float, std_dev: float) -> cp.ndarray:
        """Returns a ndarray of complex values containing results from
        a normal distribution.
        """
        return cp.random.normal(
            mean, std_dev, size=self.grid.shape
        ) + 1j * cp.random.normal(mean, std_dev, size=self.grid.shape)

    def apply_phase(self, phase: cp.ndarray, components: str = "all") -> None:
        """Applies a given phase to the specified condensate components.

        :param phase: Array containing the required phase.
        :param components: "plus", "minus", or "all". String specifying which component(s) to apply the phase to.
        """
        match components:
            case "plus":
                self.plus_component *= cp.exp(1j * phase)
            case "minus":
                self.minus_component *= cp.exp(1j * phase)
            case "all":
                self.plus_component *= cp.exp(1j * phase)
                self.minus_component *= cp.exp(1j * phase)
            case _:
                raise ValueError(f"Components type {components} is unsupported")

    def fft(self) -> None:
        """Fourier transforms real-space components and updates Fourier-space components."""
        self.fourier_plus_component = cp.fft.fftn(self.plus_component)
        self.fourier_minus_component = cp.fft.fftn(self.minus_component)

    def ifft(self) -> None:
        """Inverse Fourier transforms Fourier-space components and updates real-space components."""
        self.plus_component = cp.fft.ifftn(self.fourier_plus_component)
        self.minus_component = cp.fft.ifftn(self.fourier_minus_component)

    def density(self, components: str) -> cp.ndarray | tuple[cp.ndarray, cp.ndarray]:
        """Calculates the density of the specified component(s).

        :param components: "plus", "minus", or "all". String specifying which component(s) to
        calculate the density of.
        :return: Respective density array. If "all" is specified as the component, then both
        the plus and minus component densities are returned as a tuple, respectively.
        """
        match components.lower():
            case "plus":
                return cp.abs(self.plus_component) ** 2
            case "minus":
                return cp.abs(self.minus_component) ** 2
            case "all":
                return (
                    cp.abs(self.plus_component) ** 2,
                    cp.abs(self.minus_component) ** 2,
                )
            case _:
                raise ValueError(f"Components type {components} is unsupported")



import numpy as np
import cupy as cp


class Grid:
    """An object representing the numerical grid.
    It contains information on the number of grid points, the shape, the dimensionality, and lengths of the grid.

    :param points: Number of points in each spatial dimension.
    :type points: int or tuple of ints
    :param grid_spacings: Numerical spacing between grid points in each spatial dimension.
    :type grid_spacings: float or tuple of floats

    :ivar shape: Shape of the grid.
    :ivar ndim: Dimensionality of the grid.
    :ivar total_num_points: Total number of grid points across all dimensions.

    :ivar num_points_x: Number of points in the x-direction.
    :ivar num_points_y: (2D and 3D only) Number of points in the y-direction.
    :ivar num_points_z: (3D only) Number of points in the z-direction.
    :ivar length_x: Length of the grid in the x-direction.
    :ivar length_y: (2D and 3D only) Length of the grid in the y-direction.
    :ivar length_z: (3D only) Length of the grid in the z-direction.
    :ivar x_mesh: The x meshgrid. The dimensionality matches that of `ndim`.
    :ivar y_mesh: (2D and 3D only) The y meshgrid. The dimensionality matches that of `ndim`.
    :ivar z_mesh: (3D only) The z meshgrid. The dimensionality matches that of `ndim`.
    :ivar grid_spacing_x: Grid spacing in the x-direction.
    :ivar grid_spacing_y: (2D and 3D only) Grid spacing in the y-direction.
    :ivar grid_spacing_z: (3D only) Grid spacing in the z-direction.
    :ivar grid_spacing_product: The product of the grid spacing for each dimension.
    :ivar fourier_x_mesh: The Fourier-space x meshgrid. The dimensionality matches that of `ndim`.
    :ivar fourier_y_mesh: (2D and 3D only) The Fourier-space y meshgrid. The dimensionality matches that of `ndim`.
    :ivar fourier_z_mesh: (3D only) The Fourier-space z meshgrid. The dimensionality matches that of `ndim`.
    :ivar fourier_spacing_x: Fourier grid spacing in the x-direction.
    :ivar fourier_spacing_y: (2D and 3D only) Fourier grid spacing in the y-direction.
    :ivar fourier_spacing_z: (3D only) Fourier grid spacing in the z-direction.
    """

    def __init__(
        self, points: int | tuple[int, ...], grid_spacings: float | tuple[float, ...]
    ):
        """Constructs the grid object."""

        self.shape = points
        if isinstance(points, tuple):
            if len(points) != len(grid_spacings):
                raise ValueError(f"{points} and {grid_spacings} are not of same length")
            if len(points) > 3:
                raise ValueError(f"{points} is not a valid dimensionality")
            self.ndim = len(points)
            self.total_num_points = 1
            for point in points:
                self.total_num_points *= point
        else:
            self.ndim = 1
            self.total_num_points = points

        if self.ndim == 1:
            self._generate_1d_grids(points, grid_spacings)
        elif self.ndim == 2:
            self._generate_2d_grids(points, grid_spacings)
        elif self.ndim == 3:
            self._generate_3d_grids(points, grid_spacings)

    def _generate_1d_grids(self, points: int, grid_spacing: float):
        """Generates meshgrid for a 1D grid."""
        self.num_points_x = points
        self.grid_spacing_x = grid_spacing
        self.grid_spacing_product = self.grid_spacing_x

        self.length_x = self.num_points_x * self.grid_spacing_x
        self.x_mesh = (
            np.arange(-self.num_points_x // 2, self.num_points_x // 2)
            * self.grid_spacing_x
        )

        self.fourier_spacing_x = np.pi / (self.num_points_x // 2 * self.grid_spacing_x)
        self.fourier_x_mesh = np.fft.fftshift(
            np.arange(-self.num_points_x // 2, self.num_points_x // 2)
            * self.fourier_spacing_x
        )

        # Defined on device for use in evolution
        self.wave_number = cp.asarray(self.fourier_x_mesh**2)

    def _generate_2d_grids(
        self, points: tuple[int, ...], grid_spacings: tuple[float, ...]
    ):
        """Generates meshgrid for a 2D grid."""
        self.num_points_x, self.num_points_y = points
        self.grid_spacing_x, self.grid_spacing_y = grid_spacings
        self.grid_spacing_product = self.grid_spacing_x * self.grid_spacing_y

        self.length_x = self.num_points_x * self.grid_spacing_x
        self.length_y = self.num_points_y * self.grid_spacing_y

        x = (
            np.arange(-self.num_points_x // 2, self.num_points_x // 2)
            * self.grid_spacing_x
        )
        y = (
            np.arange(-self.num_points_y // 2, self.num_points_y // 2)
            * self.grid_spacing_y
        )
        self.x_mesh, self.y_mesh = np.meshgrid(x, y)

        # Generate Fourier space variables
        self.fourier_spacing_x = np.pi / (self.num_points_x // 2 * self.grid_spacing_x)
        self.fourier_spacing_y = np.pi / (self.num_points_y // 2 * self.grid_spacing_y)

        fourier_x = (
            np.arange(-self.num_points_x // 2, self.num_points_x // 2)
            * self.fourier_spacing_x
        )
        fourier_y = (
            np.arange(-self.num_points_y // 2, self.num_points_y // 2)
            * self.fourier_spacing_y
        )

        self.fourier_x_mesh, self.fourier_y_mesh = np.meshgrid(fourier_x, fourier_y)
        self.fourier_x_mesh = np.fft.fftshift(self.fourier_x_mesh)
        self.fourier_y_mesh = np.fft.fftshift(self.fourier_y_mesh)

        # Defined on device for use in evolution
        self.wave_number = cp.asarray(
            self.fourier_x_mesh**2 + self.fourier_y_mesh**2
        )

    def _generate_3d_grids(
        self, points: tuple[int, ...], grid_spacings: tuple[float, ...]
    ):
        """Generates meshgrid for a 3D grid."""
        self.num_points_x, self.num_points_y, self.num_points_z = points
        self.grid_spacing_x, self.grid_spacing_y, self.grid_spacing_z = grid_spacings
        self.grid_spacing_product = (
            self.grid_spacing_x * self.grid_spacing_y * self.grid_spacing_z
        )

        self.length_x = self.num_points_x * self.grid_spacing_x
        self.length_y = self.num_points_y * self.grid_spacing_y
        self.length_z = self.num_points_z * self.grid_spacing_z

        x = (
            np.arange(-self.num_points_x // 2, self.num_points_x // 2)
            * self.grid_spacing_x
        )
        y = (
            np.arange(-self.num_points_y // 2, self.num_points_y // 2)
            * self.grid_spacing_y
        )
        z = (
            np.arange(-self.num_points_z // 2, self.num_points_z // 2)
            * self.grid_spacing_z
        )
        self.x_mesh, self.y_mesh, self.z_mesh = np.meshgrid(x, y, z)

        # Generate Fourier space variables
        self.fourier_spacing_x = np.pi / (self.num_points_x // 2 * self.grid_spacing_x)
        self.fourier_spacing_y = np.pi / (self.num_points_y // 2 * self.grid_spacing_y)
        self.fourier_spacing_z = np.pi / (self.num_points_z // 2 * self.grid_spacing_z)

        fourier_x = (
            np.arange(-self.num_points_x // 2, self.num_points_x // 2)
            * self.fourier_spacing_x
        )
        fourier_y = (
            np.arange(-self.num_points_y // 2, self.num_points_y // 2)
            * self.fourier_spacing_y
        )
        fourier_z = (
            np.arange(-self.num_points_z // 2, self.num_points_z // 2)
            * self.fourier_spacing_z
        )

        self.fourier_x_mesh, self.fourier_y_mesh, self.fourier_z_mesh = np.meshgrid(
            fourier_x, fourier_y, fourier_z
        )
        self.fourier_x_mesh = np.fft.fftshift(self.fourier_x_mesh)
        self.fourier_y_mesh = np.fft.fftshift(self.fourier_y_mesh)
        self.fourier_z_mesh = np.fft.fftshift(self.fourier_z_mesh)

        # Defined on device for use in evolution
        self.wave_number = cp.asarray(
            self.fourier_x_mesh**2
            + self.fourier_y_mesh**2
            + self.fourier_z_mesh**2
        )



import cupy as cp
from pygpe.spinhalf.wavefunction import Wavefunction


def step_wavefunction(wfn: Wavefunction, params: dict) -> None:
    """Propagates the wavefunction forward one time step.

    :param wfn: The wavefunction of the system.
    :type wfn: :class:`Wavefunction`
    :param params: The parameters of the system.
    :type params: dict
    """
    _kinetic_step(wfn, params)
    wfn.ifft()
    _potential_step(wfn, params)
    wfn.fft()
    _kinetic_step(wfn, params)
    if isinstance(params["dt"], complex):
        _renormalise_wavefunction(wfn)


def _kinetic_step(wfn: Wavefunction, pm: dict) -> None:
    """Computes the kinetic energy subsystem for half a time step.

    :param wfn: The wavefunction of the system.
    :param pm: The parameters' dictionary.
    """
    wfn.fourier_plus_component *= cp.exp(-0.25 * 1j * pm["dt"] * wfn.grid.wave_number)
    wfn.fourier_minus_component *= cp.exp(-0.25 * 1j * pm["dt"] * wfn.grid.wave_number)


def _potential_step(wfn: Wavefunction, pm: dict) -> None:
    """Computes the potential subsystem for a full time step.

    :param wfn: The wavefunction of the system.
    :param pm: The parameters' dictionary.
    """
    current_plus_component = wfn.plus_component
    current_minus_component = wfn.minus_component

    wfn.plus_component *= cp.exp(
        -1j * pm["dt"] * (pm["trap"] + pm["g_plus"] * cp.abs(current_plus_component) ** 2 + pm["g_pm"] * cp.abs(
            current_minus_component))
    )
    wfn.minus_component *= cp.exp(
        -1j * pm["dt"] * (pm["trap"] + pm["g_minus"] * cp.abs(current_minus_component) ** 2 + pm["g_pm"] * cp.abs(
            current_plus_component))
    )


def _renormalise_wavefunction(wfn: Wavefunction) -> None:
    """Re-normalises the wavefunction to the correct atom number.

    :param wfn: The wavefunction of the system.
    """
    wfn.ifft()
    correct_atom_plus, correct_atom_minus = (
        wfn.atom_num_plus,
        wfn.atom_num_minus,
    )
    current_atom_plus, current_atom_minus = _calculate_atom_num(wfn)
    wfn.plus_component *= cp.sqrt(correct_atom_plus / current_atom_plus)
    wfn.minus_component *= cp.sqrt(correct_atom_minus / current_atom_minus)
    wfn.fft()


def _calculate_atom_num(wfn: Wavefunction) -> tuple[int, int]:
    """Calculates the atom number of each wavefunction component.

    :param wfn: The wavefunction of the system.
    :return: The atom numbers of the plus, zero, and minus components, respectively.
    """
    atom_num_plus = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.plus_component) ** 2
    )
    atom_num_minus = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.minus_component) ** 2
    )

    return atom_num_plus, atom_num_minus



from pygpe.shared.grid import Grid
import cupy as cp


class Wavefunction:
    """Represents the scalar BEC wavefunction.
    This class contains the wavefunction array, in addition to various useful functions for manipulating and using the
    wavefunction.

    :param grid: The numerical grid.
    :type grid: :class:`Grid`

    :ivar component: The real-space wavefunction array.
    :ivar fourier_component: The Fourier-space wavefunction array.
    :ivar atom_num: The atom number of the condensate.
    :ivar grid: Reference to the grid object of the simulation.
    """

    def __init__(self, grid: Grid):
        """Constructs the wavefunction object."""
        self.grid = grid

        self.component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_component = cp.empty(
            grid.shape, dtype="complex128"
        )  # Fourier component

        self.atom_num = 0

    def set_wavefunction(self, wavefunction: cp.ndarray) -> None:
        """Sets the wavefunction to the specified state.

        :param wavefunction:  The array to set the wavefunction as.
        :type wavefunction: `cupy.ndarray`
        """
        self.component = wavefunction
        self._update_atom_number()

    def add_noise(self, mean: float, std_dev: float) -> None:
        """Adds noise to the wavefunction using a normal distribution.

        :param mean: The mean of the normal distribution.
        :type mean: float
        :param std_dev: The standard deviation of the normal distribution.
        :type std_dev: float
        """
        self.component += self._generate_complex_normal_dist(mean, std_dev)
        self._update_atom_number()

    def _generate_complex_normal_dist(self, mean: float, std_dev: float) -> cp.ndarray:
        """Returns a ndarray of complex values containing results from
        a normal distribution.
        """
        return cp.random.normal(
            mean, std_dev, size=self.grid.shape
        ) + 1j * cp.random.normal(mean, std_dev, size=self.grid.shape)

    def apply_phase(self, phase: cp.ndarray) -> None:
        """Applies a phase to the wavefunction.

        :param phase: The phase to apply.
        :type phase: `cupy.ndarray`
        """
        self.component *= cp.exp(1j * phase)

    def _update_atom_number(self) -> None:
        self.atom_num = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.component) ** 2
        )

    def fft(self) -> None:
        """Fourier transforms real-space component and updates Fourier-space component."""
        self.fourier_component = cp.fft.fftn(self.component)

    def ifft(self) -> None:
        """Inverse Fourier transforms Fourier-space component and updates real-space component."""
        self.component = cp.fft.ifftn(self.fourier_component)

    def density(self) -> cp.ndarray:
        """

        :return: An array of the condensate density.
        :rtype: `cupy.ndarray`
        """
        return cp.abs(self.component) ** 2



"""
This file contains string constants that specify the paths of data stored in the DataManager classes.
You can modify any of these constants if you wish to change the specified path within the data HDF5 files.
"""

# Grid points
GRID_NX = "grid/nx"
GRID_NY = "grid/ny"
GRID_NZ = "grid/nz"

# Grid spacings
GRID_DX = "grid/dx"
GRID_DY = "grid/dy"
GRID_DZ = "grid/dz"

# Parameters root
PARAMETERS = "parameters"

# Scalar wavefunction
SCALAR_WAVEFUNCTION = "wavefunction"

# Spin-1/2 wavefunction
SPINHALF_WAVEFUNCTION_PLUS = "wavefunction/psi_plus"
SPINHALF_WAVEFUNCTION_MINUS = "wavefunction/psi_minus"

# Spin-1 wavefunction
SPIN1_WAVEFUNCTION_PLUS = "wavefunction/psi_plus"
SPIN1_WAVEFUNCTION_ZERO = "wavefunction/psi_zero"
SPIN1_WAVEFUNCTION_MINUS = "wavefunction/psi_minus"

# Spin-2 wavefunction
SPIN2_WAVEFUNCTION_PLUS_TWO = "wavefunction/psi_plus2"
SPIN2_WAVEFUNCTION_PLUS_ONE = "wavefunction/psi_plus1"
SPIN2_WAVEFUNCTION_ZERO = "wavefunction/psi_zero"
SPIN2_WAVEFUNCTION_MINUS_ONE = "wavefunction/psi_minus1"
SPIN2_WAVEFUNCTION_MINUS_TWO = "wavefunction/psi_minus2"



from pygpe.shared.grid import Grid
import cupy as cp


class Wavefunction:
    """Represents the spin-1 BEC wavefunction.
    This class contains the wavefunction arrays, in addition to various useful functions for manipulating and using the
    wavefunction.

    :param grid: The numerical grid.
    :type grid: :class:`Grid`

    :ivar plus_component: The real-space plus component array.
    :ivar zero_component: The real-space zero component array.
    :ivar minus_component: The real-space minus component array.
    :ivar fourier_plus_component: The Fourier-space plus component array.
    :ivar fourier_zero_component: The Fourier-space zero component array.
    :ivar fourier_minus_component: The Fourier-space minus component array.
    :ivar atom_num_plus: The atom number of the plus component.
    :ivar atom_num_zero: The atom number of the zero component.
    :ivar atom_num_minus: The atom number of the minus component.
    :ivar grid: A reference to the grid object of the simulation.
    """

    def __init__(self, grid: Grid):
        """Constructs the wavefunction object."""
        self.grid = grid

        self.plus_component = cp.empty(grid.shape, dtype="complex128")
        self.zero_component = cp.empty(grid.shape, dtype="complex128")
        self.minus_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_plus_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_zero_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_minus_component = cp.empty(grid.shape, dtype="complex128")

        self.atom_num_plus = 0
        self.atom_num_zero = 0
        self.atom_num_minus = 0

    def set_ground_state(self, ground_state: str, params: dict) -> None:
        """Sets the components of the wavefunction according to
        the ground state we wish to be in.

        :param ground_state: "polar", "ferromagnetic", or "antiferromagnetic". The ground state of the wavefunction.
        :param params: Dictionary containing condensate parameters.
        """
        ground_states = {
            "polar": _polar_initial_state,
            "ferromagnetic": _ferromagnetic_initial_state,
            "antiferromagnetic": _antiferromagnetic_initial_state,
            "BA": _broken_axisymmetry_initial_state,
        }

        ground_states[ground_state](self, params)

        self._update_atom_numbers()

    def set_custom_components(
        self,
        plus_component: cp.ndarray = None,
        zero_component: cp.ndarray = None,
        minus_component: cp.ndarray = None,
    ) -> None:
        """Sets the wavefunction components to the specified arrays.

        :param plus_component: Plus component of the wavefunction.
        :param zero_component: Zero component of the wavefunction.
        :param minus_component: Minus component of the wavefunction.
        """
        if plus_component is not None:
            self.plus_component = plus_component
        if zero_component is not None:
            self.zero_component = zero_component
        if minus_component is not None:
            self.minus_component = minus_component

    def add_noise_to_components(
        self, components: str | list[str], mean: float, std_dev: float
    ) -> None:
        """Adds noise to the specified wavefunction components
        using a normal distribution.

        :param components: "outer" or "all": The components to add noise to.
        :param mean: The mean of the normal distribution.
        :param std_dev: The standard deviation of the normal distribution.
        """

        if components.lower() == "outer":
            self.plus_component += self._generate_complex_normal_dist(mean, std_dev)
            self.minus_component += self._generate_complex_normal_dist(mean, std_dev)
        elif components.lower() == "all":
            self.plus_component += self._generate_complex_normal_dist(mean, std_dev)
            self.zero_component += self._generate_complex_normal_dist(mean, std_dev)
            self.minus_component += self._generate_complex_normal_dist(mean, std_dev)
        elif components.lower() == "middle":
            self.zero_component += self._generate_complex_normal_dist(mean, std_dev)
        else:
            raise ValueError(f"{components} is not a supported configuration")

        self._update_atom_numbers()

    def _generate_complex_normal_dist(self, mean: float, std_dev: float) -> cp.ndarray:
        """Returns a ndarray of complex values containing results from
        a normal distribution.
        """
        return cp.random.normal(
            mean, std_dev, size=self.grid.shape
        ) + 1j * cp.random.normal(mean, std_dev, size=self.grid.shape)

    def apply_phase(
        self, phase: cp.ndarray, components: str | list[str] = "all"
    ) -> None:
        """Applies a phase to specified components.

        :param phase: The phase to be applied.
        :param components: "all", "plus", "zero", "minus" or a list of strings specifying the required components.
        """
        if isinstance(components, list):
            for component in components:
                self._apply_phase_to_component(phase, component)
        elif components == "all":
            for component in ["plus", "zero", "minus"]:
                self._apply_phase_to_component(phase, component)
        elif isinstance(components, str):
            self._apply_phase_to_component(phase, components)
        else:
            raise ValueError(f"Components type {components} is unsupported")

    def _apply_phase_to_component(self, phase: cp.ndarray, component: str) -> None:
        if component == "plus":
            self.plus_component *= cp.exp(1j * phase)
        elif component == "zero":
            self.zero_component *= cp.exp(1j * phase)
        elif component == "minus":
            self.minus_component *= cp.exp(1j * phase)
        else:
            raise ValueError(f"Component type {component} is unsupported")

    def _update_atom_numbers(self) -> None:
        self.atom_num_plus = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.plus_component) ** 2
        )
        self.atom_num_zero = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.zero_component) ** 2
        )
        self.atom_num_minus = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.minus_component) ** 2
        )

    def fft(self) -> None:
        """Fourier transforms real-space components and updates Fourier-space components."""
        self.fourier_plus_component = cp.fft.fftn(self.plus_component)
        self.fourier_zero_component = cp.fft.fftn(self.zero_component)
        self.fourier_minus_component = cp.fft.fftn(self.minus_component)

    def ifft(self) -> None:
        """Inverse Fourier transforms Fourier-space components and updates real-space components."""
        self.plus_component = cp.fft.ifftn(self.fourier_plus_component)
        self.zero_component = cp.fft.ifftn(self.fourier_zero_component)
        self.minus_component = cp.fft.ifftn(self.fourier_minus_component)

    def density(self) -> cp.ndarray:
        """Returns an array of the total condensate density.

        :return: Total condensate density.
        """
        return (
            cp.abs(self.plus_component) ** 2
            + cp.abs(self.zero_component) ** 2
            + cp.abs(self.minus_component) ** 2
        )


def _polar_initial_state(wfn: Wavefunction, params: dict) -> None:
    """Sets wavefunction components to (easy-axis) polar state."""
    wfn.plus_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.zero_component = cp.sqrt(params["n0"]) * cp.ones(
        wfn.grid.shape, dtype="complex128"
    )
    wfn.minus_component = cp.zeros(wfn.grid.shape, dtype="complex128")


def _ferromagnetic_initial_state(wfn: Wavefunction, params: dict) -> None:
    """Sets wavefunction components to ferromagnetic state."""
    wfn.plus_component = cp.sqrt(params["n0"]) * cp.ones(
        wfn.grid.shape, dtype="complex128"
    )
    wfn.zero_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus_component = cp.zeros(wfn.grid.shape, dtype="complex128")


def _antiferromagnetic_initial_state(wfn: Wavefunction, params: dict) -> None:
    """Sets wavefunction components to antiferromagnetic state."""
    p = params["p"]  # Linear Zeeman
    c2 = params["c2"]  # Spin-dependent interaction strength
    n = params["n0"]

    wfn.plus_component = (
        cp.sqrt(n)
        * cp.sqrt((1 + p / c2) / 2)
        * cp.ones(wfn.grid.shape, dtype="complex128")
    )
    wfn.zero_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus_component = (
        cp.sqrt(n)
        * cp.sqrt((1 - p / c2) / 2)
        * cp.ones(wfn.grid.shape, dtype="complex128")
    )


def _broken_axisymmetry_initial_state(wfn: Wavefunction, params: dict) -> None:
    """Sets wavefunction components to antiferromagnetic state."""
    p = params["p"]  # Linear Zeeman
    q = params["q"]  # Quadratic Zeeman
    c2 = params["c2"]  # Spin-dependent interaction strength
    n = params["n0"]

    wfn.plus_component = (
        cp.sqrt(n)
        * (q + p)
        / (2 * q)
        * cp.sqrt((-(p**2) + q**2 + 2 * c2 * n * q) / (2 * c2 * n * q))
        * cp.ones(wfn.grid.shape, dtype="complex128")
    )
    wfn.zero_component = (
        cp.sqrt(n)
        * cp.sqrt(
            (q**2 - p**2)
            * (-(p**2) - q**2 + 2 * c2 * n * q)
            / (4 * c2 * n * q**3)
        )
        * cp.ones(wfn.grid.shape, dtype="complex128")
    )
    wfn.minus_component = (
        cp.sqrt(n)
        * (q - p)
        / (2 * q)
        * cp.sqrt((-(p**2) + q**2 + 2 * c2 * n * q) / (2 * c2 * n * q))
        * cp.ones(wfn.grid.shape, dtype="complex128")
    )



import cupy as cp
from pygpe.scalar.wavefunction import Wavefunction


def step_wavefunction(wfn: Wavefunction, params: dict) -> None:
    """Propagates the wavefunction forward one time step.

    :param wfn: The wavefunction of the system.
    :type wfn: :class:`Wavefunction`
    :param params: The parameters of the system.
    :type params: dict
    """
    _kinetic_step(wfn, params)
    wfn.ifft()
    _potential_step(wfn, params)
    wfn.fft()
    _kinetic_step(wfn, params)
    if isinstance(params["dt"], complex):
        _renormalise_wavefunction(wfn)


def _kinetic_step(wfn: Wavefunction, pm: dict) -> None:
    """Computes the kinetic energy subsystem for half a time step.

    :param wfn: The wavefunction of the system.
    :param pm: The parameters' dictionary.
    """
    wfn.fourier_component *= cp.exp(-0.25 * 1j * pm["dt"] * wfn.grid.wave_number)


def _potential_step(wfn: Wavefunction, pm: dict) -> None:
    """Computes the potential subsystem for a full time step.

    :param wfn: The wavefunction of the system.
    :param pm: The parameters' dictionary.
    """
    wfn.component *= cp.exp(
        -1j * pm["dt"] * (pm["trap"] + pm["g"] * cp.abs(wfn.component) ** 2)
    )


def _renormalise_wavefunction(wfn: Wavefunction) -> None:
    """Re-normalises the wavefunction to the correct atom number.

    :param wfn: The wavefunction of the system.
    """
    wfn.ifft()
    correct_atom_num = wfn.atom_num
    current_atom_num = _calculate_atom_num(wfn)
    wfn.component *= cp.sqrt(correct_atom_num / current_atom_num)
    wfn.fft()


def _calculate_atom_num(wfn: Wavefunction) -> float:
    """Calculates the current atom number of the wavefunction.

    :param wfn: The wavefunction of the system.
    :return: The atom number.
    """
    return wfn.grid.grid_spacing_product * cp.sum(cp.abs(wfn.component) ** 2)



import numpy as np
from pygpe.shared.grid import Grid


def _generate_positions(grid: Grid, num_vortices: int, threshold: float) -> iter:
    """Generates and returns a list of positions that are separated by at least
    `threshold`.
    """
    print(f"Attempting to find {num_vortices} positions...")
    max_iter = 10000
    vortex_positions = []

    iterations = 0
    while len(vortex_positions) < num_vortices:
        if iterations > max_iter:
            print(
                f"WARNING: Number of iterations exceeded maximum, "
                f"returning with only {len(vortex_positions)} positions\n"
            )
            return vortex_positions

        position = np.random.uniform(
            -grid.length_x / 2, grid.length_x / 2
        ), np.random.uniform(-grid.length_y / 2, grid.length_y / 2)

        if _position_sufficiently_far(position, vortex_positions, threshold):
            vortex_positions.append(position)

        iterations += 1

    print(f"Successfully found {num_vortices} positions in {iterations} iterations!")
    return iter(vortex_positions)


def _position_sufficiently_far(
    position: tuple, accepted_positions: list[tuple], threshold: float
) -> bool:
    """Tests that the given `position` is at least `threshold` away from all the positions
    currently in `accepted_positions`.
    """
    # Special case where accepted_positions is empty
    if not accepted_positions:
        return True

    for accepted_pos in accepted_positions:
        if abs(position[0] - accepted_pos[0]) > threshold:
            if abs(position[1] - accepted_pos[1]) > threshold:
                return True
    return False


def _heaviside(array: np.ndarray) -> np.ndarray:
    """Computes the heaviside function on a given array and returns the result."""
    return np.where(array < 0, np.zeros(array.shape), np.ones(array.shape))


def vortex_phase_profile(grid: Grid, num_vortices: int, threshold: float) -> np.ndarray:
    """Constructs a 2D phase profile consisting of 2pi phase windings.
    This phase can be applied to a wavefunction to generate different types of vortices.

    :param grid: The 2D grid of the system.
    :type grid: :class:`Grid`
    :param num_vortices: The total number of vortices to be included in the phase profile.
    :type num_vortices: int
    :param threshold: The minimum distance allowed between any two vortices.
    :type threshold: float
    """
    vortex_positions_iter = _generate_positions(grid, num_vortices, threshold)

    phase = np.zeros((grid.num_points_x, grid.num_points_y), dtype="float32")

    for _ in range(num_vortices // 2):
        phase_temp = np.zeros((grid.num_points_x, grid.num_points_y), dtype="float32")
        x_pos_minus, y_pos_minus = next(
            vortex_positions_iter
        )  # Negative circulation vortex
        x_pos_plus, y_pos_plus = next(
            vortex_positions_iter
        )  # Positive circulation vortex

        # Aux variables
        y_minus = 2 * np.pi / grid.length_y * (grid.y_mesh - y_pos_minus)
        x_minus = 2 * np.pi / grid.length_x * (grid.x_mesh - x_pos_minus)
        y_plus = 2 * np.pi / grid.length_y * (grid.y_mesh - y_pos_plus)
        x_plus = 2 * np.pi / grid.length_x * (grid.x_mesh - x_pos_plus)

        heaviside_x_plus = _heaviside(x_plus)
        heaviside_x_minus = _heaviside(x_minus)

        for nn in np.arange(-5, 6):
            phase_temp += (
                np.arctan(
                    np.tanh((y_minus + 2 * np.pi * nn) / 2)
                    * np.tan((x_minus - np.pi) / 2)
                )
                - np.arctan(
                    np.tanh((y_plus + 2 * np.pi * nn) / 2)
                    * np.tan((x_plus - np.pi) / 2)
                )
                + np.pi * (heaviside_x_plus - heaviside_x_minus)
            )
        phase_temp -= (
            2
            * np.pi
            * (grid.y_mesh - grid.y_mesh.min())
            * (x_pos_plus - x_pos_minus)
            / (grid.length_y * grid.length_x)
        )
        phase += phase_temp

    return phase



import cupy as cp
from pygpe.spintwo.wavefunction import Wavefunction


def step_wavefunction(wfn: Wavefunction, params: dict) -> None:
    """Propagates the wavefunction forward one time step.

    :param wfn: The wavefunction of the system.
    :type wfn: :class:`Wavefunction`
    :param params: The parameters of the system.
    :type params: dict
    """
    _kinetic_step(wfn, params)
    wfn.ifft()
    _interaction_step(wfn, params)
    wfn.fft()
    _kinetic_step(wfn, params)
    if isinstance(params["dt"], complex):
        _renormalise_wavefunction(wfn)


def _kinetic_step(wfn: Wavefunction, pm: dict) -> None:
    """Computes the kinetic energy subsystem for half a time step.

    :param wfn: The wavefunction of the system.
    :param pm:  The parameters' dictionary.
    """
    wfn.fourier_plus2_component *= cp.exp(-0.25 * 1j * pm["dt"] * wfn.grid.wave_number)
    wfn.fourier_plus1_component *= cp.exp(-0.25 * 1j * pm["dt"] * wfn.grid.wave_number)
    wfn.fourier_zero_component *= cp.exp(-0.25 * 1j * pm["dt"] * wfn.grid.wave_number)
    wfn.fourier_minus1_component *= cp.exp(-0.25 * 1j * pm["dt"] * wfn.grid.wave_number)
    wfn.fourier_minus2_component *= cp.exp(-0.25 * 1j * pm["dt"] * wfn.grid.wave_number)


def _interaction_step(wfn: Wavefunction, pm: dict) -> None:
    # Calculate density and singlets
    n = _density(wfn)
    a20 = _singlet_duo(wfn)

    # Perform singlet step
    temp_wfn = _evolve_spin_singlet(wfn, n, a20, pm)

    # Calculate spin
    fp, fz = _calculate_spin_vectors(temp_wfn)
    mod_f = cp.sqrt(fz**2 + abs(fp) ** 2)

    # Evolve spin term c2 * F^2
    qfactor, q2factor, q3factor, q4factor = _calc_q_factors(mod_f, pm)

    fzq = cp.nan_to_num(fz / mod_f)
    fpq = cp.nan_to_num(fp / mod_f)

    qpsi = _calc_qpsi(fzq, fpq, temp_wfn)
    q2psi = _calc_qpsi(fzq, fpq, qpsi)
    q3psi = _calc_qpsi(fzq, fpq, q2psi)
    q4psi = _calc_qpsi(fzq, fpq, q3psi)

    for ii in range(len(temp_wfn)):
        temp_wfn[ii] += (
            qfactor * qpsi[ii]
            + q2factor * q2psi[ii]
            + q3factor * q3psi[ii]
            + q4factor * q4psi[ii]
        )

    # Evolve (c0+c4)*n + (V + pm + qm^2):
    for ii in range(len(temp_wfn)):
        m_f = 2 - ii  # Current spin component
        temp_wfn[ii] *= cp.exp(
            -1j
            * pm["dt"]
            * (
                (pm["c0"] + pm["c4"]) * n
                + pm["trap"]
                - pm["p"] * m_f
                + pm["q"] * m_f**2
            )
        )

    # Update wavefunction arrays
    wfn.plus2_component = temp_wfn[0]
    wfn.plus1_component = temp_wfn[1]
    wfn.zero_component = temp_wfn[2]
    wfn.minus1_component = temp_wfn[3]
    wfn.minus2_component = temp_wfn[4]


def _density(wfn: Wavefunction) -> cp.ndarray:
    return (
        abs(wfn.plus2_component) ** 2
        + abs(wfn.plus1_component) ** 2
        + abs(wfn.zero_component) ** 2
        + abs(wfn.minus1_component) ** 2
        + abs(wfn.minus2_component) ** 2
    )


def _singlet_duo(wfn: Wavefunction) -> cp.ndarray:
    return (
        1
        / cp.sqrt(5)
        * (
            wfn.zero_component**2
            - 2 * wfn.plus1_component * wfn.minus1_component
            + 2 * wfn.plus2_component * wfn.minus2_component
        )
    )


def _evolve_spin_singlet(
    wfn: Wavefunction, dens: cp.ndarray, singlet: cp.ndarray, pm: dict
) -> list[cp.ndarray]:
    s = cp.nan_to_num(cp.sqrt(dens**2 - abs(singlet) ** 2))
    cos_term = cp.cos(pm["c4"] * s * pm["dt"])
    sin_term = cp.sin(pm["c4"] * s * pm["dt"]) / s
    sin_term[s == 0] = 0  # Corrects division by 0

    psi_p2 = (
        wfn.plus2_component * cos_term
        + 1j
        * (dens * wfn.plus2_component - singlet * cp.conj(wfn.minus2_component))
        * sin_term
    )
    psi_p1 = (
        wfn.plus1_component * cos_term
        + 1j
        * (dens * wfn.plus1_component + singlet * cp.conj(wfn.minus1_component))
        * sin_term
    )
    psi_0 = (
        wfn.zero_component * cos_term
        + 1j
        * (dens * wfn.zero_component - singlet * cp.conj(wfn.zero_component))
        * sin_term
    )
    psi_m1 = (
        wfn.minus1_component * cos_term
        + 1j
        * (dens * wfn.minus1_component + singlet * cp.conj(wfn.plus1_component))
        * sin_term
    )
    psi_m2 = (
        wfn.minus2_component * cos_term
        + 1j
        * (dens * wfn.minus2_component - singlet * cp.conj(wfn.plus2_component))
        * sin_term
    )

    return [psi_p2, psi_p1, psi_0, psi_m1, psi_m2]


def _calculate_spin_vectors(wfn: list[cp.ndarray]):
    fp = cp.sqrt(6) * (wfn[1] * cp.conj(wfn[2]) + wfn[2] * cp.conj(wfn[3])) + 2 * (
        wfn[3] * cp.conj(wfn[4]) + wfn[0] * cp.conj(wfn[1])
    )
    fz = 2 * (abs(wfn[0]) ** 2 - abs(wfn[4]) ** 2) + abs(wfn[1]) ** 2 - abs(wfn[3]) ** 2
    return fp, fz


def _calc_q_factors(mod_f: cp.ndarray, pm: dict):
    cos1, sin1 = cp.cos(pm["c2"] * mod_f * pm["dt"]), cp.sin(
        pm["c2"] * mod_f * pm["dt"]
    )
    cos2, sin2 = cp.cos(2 * pm["c2"] * mod_f * pm["dt"]), cp.sin(
        2 * pm["c2"] * mod_f * pm["dt"]
    )
    qfactor = 1j * (-4 / 3 * sin1 + 1 / 6 * sin2)
    q2factor = -5 / 4 + 4 / 3 * cos1 - 1 / 12 * cos2
    q3factor = 1j * (1 / 3 * sin1 - 1 / 6 * sin2)
    q4factor = 1 / 4 - 1 / 3 * cos1 + 1 / 12 * cos2

    return qfactor, q2factor, q3factor, q4factor


def _calc_qpsi(fz, fp, wfn):
    qpsi = [
        2 * fz * wfn[0] + fp * wfn[1],
        cp.conj(fp) * wfn[0] + fz * wfn[1] + cp.sqrt(3 / 2) * fp * wfn[2],
        cp.sqrt(3 / 2) * (cp.conj(fp) * wfn[1] + fp * wfn[3]),
        cp.sqrt(3 / 2) * cp.conj(fp) * wfn[2] - fz * wfn[3] + fp * wfn[4],
        cp.conj(fp) * wfn[3] - 2 * fz * wfn[4],
    ]

    return qpsi


def _renormalise_wavefunction(wfn: Wavefunction) -> None:
    """Re-normalises the wavefunction to the correct atom number.

    :param wfn: The wavefunction of the system.
    """
    wfn.ifft()
    correct_atom_plus2 = wfn.atom_num_plus2
    correct_atom_plus1 = wfn.atom_num_plus1
    correct_atom_zero = wfn.atom_num_zero
    correct_atom_minus1 = wfn.atom_num_minus1
    correct_atom_minus2 = wfn.atom_num_minus2

    (
        current_atom_plus2,
        current_atom_plus1,
        current_atom_zero,
        current_atom_minus1,
        current_atom_minus2,
    ) = _calculate_atom_num(wfn)

    wfn.plus2_component *= cp.sqrt(correct_atom_plus2 / current_atom_plus2)
    wfn.plus1_component *= cp.sqrt(correct_atom_plus1 / current_atom_plus1)
    wfn.zero_component *= cp.sqrt(correct_atom_zero / current_atom_zero)
    wfn.minus1_component *= cp.sqrt(correct_atom_minus1 / current_atom_minus1)
    wfn.minus2_component *= cp.sqrt(correct_atom_minus2 / current_atom_minus2)
    wfn.fft()


def _calculate_atom_num(wfn: Wavefunction) -> tuple[int, int, int, int, int]:
    """Calculates the atom number of each wavefunction component.

    :param wfn: The wavefunction of the system.
    :return: The atom numbers of the plus two, plus one, zero, minus one, and minus two
    components, respectively.
    """
    atom_num_plus2 = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.plus2_component) ** 2
    )
    atom_num_plus1 = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.plus1_component) ** 2
    )
    atom_num_zero = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.zero_component) ** 2
    )
    atom_num_minus1 = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.minus1_component) ** 2
    )
    atom_num_minus2 = wfn.grid.grid_spacing_product * cp.sum(
        cp.abs(wfn.minus2_component) ** 2
    )

    return (
        atom_num_plus2,
        atom_num_plus1,
        atom_num_zero,
        atom_num_minus1,
        atom_num_minus2,
    )



from pygpe.shared.grid import Grid
import cupy as cp


class Wavefunction:
    """Represents the spin-2 BEC wavefunction.
    This class contains the wavefunction arrays, in addition to various useful functions for manipulating and using the
    wavefunction.

    :param grid: The numerical grid.
    :type grid: :class:`Grid`

    :ivar plus2_component: The real-space +2 component array.
    :ivar plus1_component: The real-space +1 component array.
    :ivar zero_component: The real-space 0 component array.
    :ivar minus1_component: The real-space -1 component array.
    :ivar minus2_component: The real-space -2 component array.
    :ivar fourier_plus2_component: The Fourier-space +2 component array.
    :ivar fourier_plus1_component: The Fourier-space +1 component array.
    :ivar fourier_zero_component: The Fourier-space 0 component array.
    :ivar fourier_minus1_component: The Fourier-space -1 component array.
    :ivar fourier_minus2_component: The Fourier-space -2 component array.
    :ivar atom_num_plus2: The atom number of the +2 component.
    :ivar atom_num_plus1: The atom number of the +1 component.
    :ivar atom_num_zero: The atom number of the 0 component.
    :ivar atom_num_minus1: The atom number of the -1 component.
    :ivar atom_num_minus2: The atom number of the -2 component.
    :ivar grid: Reference to the grid object of the simulation.
    """

    def __init__(self, grid: Grid):
        """Constructs the wavefunction object."""
        self.grid = grid

        self.plus2_component = cp.empty(grid.shape, dtype="complex128")
        self.plus1_component = cp.empty(grid.shape, dtype="complex128")
        self.zero_component = cp.empty(grid.shape, dtype="complex128")
        self.minus1_component = cp.empty(grid.shape, dtype="complex128")
        self.minus2_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_plus2_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_plus1_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_zero_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_minus1_component = cp.empty(grid.shape, dtype="complex128")
        self.fourier_minus2_component = cp.empty(grid.shape, dtype="complex128")

        self.atom_num_plus2 = 0
        self.atom_num_plus1 = 0
        self.atom_num_zero = 0
        self.atom_num_minus1 = 0
        self.atom_num_minus2 = 0

    def set_ground_state(self, ground_state: str, params: dict) -> None:
        """Sets the components of the wavefunction according to
        the ground state we wish to be in.

        :param ground_state: "UN", "BN", "F2p", "F2m", "F1p", "F1m" or "cyclic".
        :param params: Dictionary containing condensate parameters.
        """
        ground_states = {
            "UN": _uniaxial_initial_state,
            "BN": _biaxial_initial_state,
            "F2p": _ferromagnetic2p_initial_state,
            "F2m": _ferromagnetic2m_initial_state,
            "F1p": _ferromagnetic1p_initial_state,
            "F1m": _ferromagnetic1m_initial_state,
            "cyclic": _cyclic_initial_state,
        }

        ground_states[ground_state](self, params)

        self._update_atom_numbers()

    def set_custom_components(
        self,
        plus2_component: cp.ndarray = None,
        plus1_component: cp.ndarray = None,
        zero_component: cp.ndarray = None,
        minus1_component: cp.ndarray = None,
        minus2_component: cp.ndarray = None,
    ) -> None:
        """Sets the wavefunction components to the specified arrays.

        :param plus2_component: Plus two component of the wavefunction.
        :param plus1_component: Plus one component of the wavefunction.
        :param zero_component: Zero component of the wavefunction.
        :param minus1_component: Minus one component of the wavefunction.
        :param minus2_component: Minus two component of the wavefunction.
        """
        if plus2_component is not None:
            self.plus2_component = plus2_component
        if plus1_component is not None:
            self.plus1_component = plus1_component
        if zero_component is not None:
            self.zero_component = zero_component
        if minus1_component is not None:
            self.minus1_component = minus1_component
        if minus2_component is not None:
            self.minus2_component = minus2_component

    def add_noise_to_components(
        self, components: str, mean: float, std_dev: float
    ) -> None:
        """Adds noise to the specified wavefunction components
        using a normal distribution.

        :param components: "outer", "inner", "outer+inner", "middle", or "all": The components to add noise to.
        :param mean: The mean of the normal distribution.
        :param std_dev: The standard deviation of the normal distribution.
        """
        if components.lower() == "outer":
            self.plus2_component += self._generate_complex_normal_dist(mean, std_dev)
            self.minus2_component += self._generate_complex_normal_dist(mean, std_dev)
        elif components.lower() == "inner":
            self.plus1_component += self._generate_complex_normal_dist(mean, std_dev)
            self.minus1_component += self._generate_complex_normal_dist(mean, std_dev)
        elif components.lower() == "outer+inner":
            self.plus2_component += self._generate_complex_normal_dist(mean, std_dev)
            self.plus1_component += self._generate_complex_normal_dist(mean, std_dev)
            self.minus1_component += self._generate_complex_normal_dist(mean, std_dev)
            self.minus2_component += self._generate_complex_normal_dist(mean, std_dev)
        elif components.lower() == "middle":
            self.zero_component += self._generate_complex_normal_dist(mean, std_dev)
        elif components.lower() == "all":
            self.plus2_component += self._generate_complex_normal_dist(mean, std_dev)
            self.plus1_component += self._generate_complex_normal_dist(mean, std_dev)
            self.zero_component += self._generate_complex_normal_dist(mean, std_dev)
            self.minus1_component += self._generate_complex_normal_dist(mean, std_dev)
            self.minus2_component += self._generate_complex_normal_dist(mean, std_dev)
        else:
            raise ValueError(f"{components} is not a supported configuration")

        self._update_atom_numbers()

    def _generate_complex_normal_dist(self, mean: float, std_dev: float) -> cp.ndarray:
        """Returns a ndarray of complex values containing results from
        a normal distribution.
        """
        return cp.random.normal(
            mean, std_dev, size=self.grid.shape
        ) + 1j * cp.random.normal(mean, std_dev, size=self.grid.shape)

    def apply_phase(
        self, phase: cp.ndarray, components: str | list[str] = "all"
    ) -> None:
        """Applies a phase to specified components.

        :param phase: The phase to be applied.
        :param components: "all", "plus2", "plus1", "zero", "minus1", "minus2" or a list of strings specifying the
            required components.
        """
        if isinstance(components, list):
            for component in components:
                self._apply_phase_to_component(phase, component)
        elif components == "all":
            for component in ["plus2", "plus1", "zero", "minus1", "minus2"]:
                self._apply_phase_to_component(phase, component)
        elif isinstance(components, str):
            self._apply_phase_to_component(phase, components)
        else:
            raise ValueError(f"Components type {components} is unsupported")

    def _apply_phase_to_component(self, phase: cp.ndarray, component: str):
        if component == "plus2":
            self.plus2_component *= cp.exp(1j * phase)
        elif component == "plus1":
            self.plus1_component *= cp.exp(1j * phase)
        elif component == "zero":
            self.zero_component *= cp.exp(1j * phase)
        elif component == "minus1":
            self.minus1_component *= cp.exp(1j * phase)
        elif component == "minus2":
            self.minus1_component *= cp.exp(1j * phase)
        else:
            raise ValueError(f"Component type {component} is unsupported")

    def _update_atom_numbers(self):
        self.atom_num_plus2 = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.plus2_component) ** 2
        )
        self.atom_num_plus1 = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.plus1_component) ** 2
        )
        self.atom_num_zero = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.zero_component) ** 2
        )
        self.atom_num_minus1 = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.minus1_component) ** 2
        )
        self.atom_num_minus2 = self.grid.grid_spacing_product * cp.sum(
            cp.abs(self.minus2_component) ** 2
        )

    def fft(self):
        """Fourier transforms real-space components and updates Fourier-space components."""
        self.fourier_plus2_component = cp.fft.fftn(self.plus2_component)
        self.fourier_plus1_component = cp.fft.fftn(self.plus1_component)
        self.fourier_zero_component = cp.fft.fftn(self.zero_component)
        self.fourier_minus1_component = cp.fft.fftn(self.minus1_component)
        self.fourier_minus2_component = cp.fft.fftn(self.minus2_component)

    def ifft(self):
        """Inverse Fourier transforms Fourier-space components and updates real-space components."""
        self.plus2_component = cp.fft.ifftn(self.fourier_plus2_component)
        self.plus1_component = cp.fft.ifftn(self.fourier_plus1_component)
        self.zero_component = cp.fft.ifftn(self.fourier_zero_component)
        self.minus1_component = cp.fft.ifftn(self.fourier_minus1_component)
        self.minus2_component = cp.fft.ifftn(self.fourier_minus2_component)


def _uniaxial_initial_state(wfn: Wavefunction, params: dict):
    """Sets wavefunction components to uniaxial nematic state."""
    wfn.plus2_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.plus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.zero_component = cp.sqrt(params["n0"]) * cp.ones(
        wfn.grid.shape, dtype="complex128"
    )
    wfn.minus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus2_component = cp.zeros(wfn.grid.shape, dtype="complex128")


def _biaxial_initial_state(wfn: Wavefunction, params: dict):
    """Sets wavefunction components to biaxial nematic polar state."""
    wfn.plus2_component = (
        cp.sqrt(params["n0"])
        / cp.sqrt(2.0)
        * cp.ones(wfn.grid.shape, dtype="complex128")
    )
    wfn.plus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.zero_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus2_component = (
        cp.sqrt(params["n0"])
        / cp.sqrt(2.0)
        * cp.ones(wfn.grid.shape, dtype="complex128")
    )


def _ferromagnetic2p_initial_state(wfn: Wavefunction, params: dict):
    """Sets wavefunction components to ferromagnetic (F=2) state, with atoms in the plus two component."""
    wfn.plus2_component = cp.sqrt(params["n0"]) * cp.ones(
        wfn.grid.shape, dtype="complex128"
    )
    wfn.plus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.zero_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus2_component = cp.zeros(wfn.grid.shape, dtype="complex128")


def _ferromagnetic2m_initial_state(wfn: Wavefunction, params: dict):
    """Sets wavefunction components to ferromagnetic (F=2) state, with atoms in the minus two component."""
    wfn.plus2_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.plus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.zero_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus2_component = cp.sqrt(params["n0"]) * cp.ones(
        wfn.grid.shape, dtype="complex128"
    )


def _ferromagnetic1p_initial_state(wfn: Wavefunction, params: dict):
    """Sets wavefunction components to ferromagnetic (F=1) state, with atoms in the plus one component."""
    wfn.plus2_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.plus1_component = cp.sqrt(params["n0"]) * cp.ones(
        wfn.grid.shape, dtype="complex128"
    )
    wfn.zero_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus2_component = cp.zeros(wfn.grid.shape, dtype="complex128")


def _ferromagnetic1m_initial_state(wfn: Wavefunction, params: dict):
    """Sets wavefunction components to ferromagnetic (F=1) state, with atoms in the minus one component."""
    wfn.plus2_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.plus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.zero_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus1_component = cp.sqrt(params["n0"]) * cp.ones(
        wfn.grid.shape, dtype="complex128"
    )
    wfn.minus2_component = cp.zeros(wfn.grid.shape, dtype="complex128")


def _cyclic_initial_state(wfn: Wavefunction, params: dict):
    """Sets wavefunction components to (two-component) cyclic state,
    with atoms in the plus two and minus one components."""
    fz = params["p"] + params["q"] / (params["c2"] * params["n0"])

    wfn.plus2_component = (
        cp.sqrt(params["n0"])
        * cp.sqrt((1 + fz) / 3)
        * cp.ones(wfn.grid.shape, dtype="complex128")
    )
    wfn.plus1_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.zero_component = cp.zeros(wfn.grid.shape, dtype="complex128")
    wfn.minus1_component = (
        cp.sqrt(params["n0"])
        * cp.sqrt((2 - fz) / 3)
        * cp.ones(wfn.grid.shape, dtype="complex128")
    )
    wfn.minus2_component = cp.zeros(wfn.grid.shape, dtype="complex128")
